//! Data structures for representing the [asciicast v2] format
//!
//! [asciicast v2]: https://github.com/asciinema/asciinema/blob/develop/doc/asciicast-v2.md

use std::{collections::HashMap, fmt, io::Cursor};

use serde::{
  de::{
    Deserialize, Deserializer, Error as DeError, SeqAccess, Unexpected, Visitor,
  },
  ser::{Serialize, SerializeSeq, Serializer},
};
use serde_bytes::{ByteBuf, Bytes};
use serde_json::ser::{CharEscape, CompactFormatter, Formatter};

use crate::char_escaper::CharEscaper;

#[derive(Clone, Debug, Builder, Serialize, Deserialize)]
pub struct Header {
  #[builder(setter(skip))]
  pub version: Version,
  pub width: u32,
  pub height: u32,

  #[builder(setter(strip_option), default)]
  pub timestamp: Option<u32>,
  #[builder(setter(strip_option), default)]
  pub duration: Option<f64>,
  #[builder(setter(strip_option), default)]
  pub idle_time_limit: Option<f64>,
  #[builder(setter(into, strip_option), default)]
  pub command: Option<String>,
  #[builder(setter(into, strip_option), default)]
  pub title: Option<String>,
  #[builder(setter(into, strip_option), default)]
  pub env: Option<HashMap<String, String>>,
  #[builder(setter(strip_option), default)]
  pub theme: Option<Theme>,
}

#[derive(Debug, Clone, Copy, Serialize, Deserialize)]
pub struct Version(u32);

impl Default for Version {
  fn default() -> Self {
    Version(2)
  }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Theme {
  fg: String,
  bg: String,
  palette: String,
}

#[derive(Debug, Clone, PartialEq)]
pub struct Event(
  /// Elapsed time
  pub f64,
  pub EventKind,
);

#[derive(Debug, Clone, Serialize, Deserialize, PartialEq)]
pub enum EventKind {
  Output(Vec<u8>),
  Input(Vec<u8>),
}

impl Serialize for Event {
  fn serialize<S: Serializer>(
    &self,
    s: S,
  ) -> std::result::Result<S::Ok, S::Error> {
    let mut seq = s.serialize_seq(Some(3))?;
    seq.serialize_element(&self.0)?;
    match &self.1 {
      EventKind::Output(s) => {
        seq.serialize_element("o")?;
        let ce = CharEscaper::new(s);
        let s = ce.to_string();
        seq.serialize_element(&s)?;
      }
      EventKind::Input(s) => {
        seq.serialize_element("i")?;
        let ce = CharEscaper::new(s);
        let s = ce.to_string();
        seq.serialize_element(&s)?;
      }
    }
    seq.end()
  }
}

impl<'d> Deserialize<'d> for Event {
  fn deserialize<D>(de: D) -> std::result::Result<Self, D::Error>
  where
    D: Deserializer<'d>,
  {
    struct Vis;
    impl<'de> Visitor<'de> for Vis {
      type Value = Event;

      fn expecting(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.write_str("an array of [timestamp, o or i, data]")
      }

      fn visit_seq<A>(self, mut seq: A) -> Result<Self::Value, A::Error>
      where
        A: SeqAccess<'de>,
      {
        let elapsed: f64 = match seq.next_element()? {
          Some(v) => v,
          None => {
            return Err(A::Error::invalid_length(0, &"an array of length 3"))
          }
        };

        let io: char = match seq.next_element()? {
          Some(v) => v,
          None => {
            return Err(A::Error::invalid_length(1, &"an array of length 3"))
          }
        };

        let data: &'de [u8] = match seq.next_element()? {
          Some(v) => v,
          None => {
            return Err(A::Error::invalid_length(2, &"an array of length 3"))
          }
        };
        let data = data.to_vec();

        let event_kind = match io {
          'i' => EventKind::Input(data),
          'o' => EventKind::Output(data),
          _ => {
            return Err(A::Error::invalid_value(
              Unexpected::Char(io),
              &"either i or o",
            ))
          }
        };

        Ok(Event(elapsed, event_kind))
      }
    }

    de.deserialize_seq(Vis)
  }
}

#[test]
fn test() {
  let evt = Event(1.5, EventKind::Output(vec![69, 42, 255, 97, 240, 159, 135, 122].into()));
  let evt_ser = serde_json::to_string(&evt).unwrap();
  eprintln!("ser: {evt_ser}");
  let evt_de: Event = serde_json::from_str(&evt_ser).unwrap();
  assert_eq!(evt, evt_de);
  panic!()
}
