use std::env;
use std::ffi::CString;
use std::io::stdin;
use std::os::unix::prelude::{AsRawFd, OsStrExt};
use std::process::Command;
use std::sync::mpsc::{self, Receiver, Sender};
use std::time::Instant;

use anyhow::{Context, Result};
use libc::{STDERR_FILENO, STDIN_FILENO, STDOUT_FILENO};
use nix::fcntl::{open, OFlag};
use nix::pty::openpty;
use nix::sys::select::{select, FdSet};
use nix::sys::stat::Mode;
use nix::unistd::{
  close, dup, dup2, execvpe, fork, fsync, read, setsid, ttyname, write,
  ForkResult,
};

use crate::asciicast::{Event, EventKind};
use crate::client::raw_term::RawTerm;

pub struct Terminal {
  event_tx: Sender<Event>,
  command: Command,
}

impl Terminal {
  pub fn setup(command: Command) -> Result<(Self, Receiver<Event>)> {
    // Set up channels
    let (event_tx, event_rx) = mpsc::channel::<Event>();

    let term = Terminal { event_tx, command };
    Ok((term, event_rx))
  }

  pub fn wait_until_complete(self) -> Result<()> {
    self.run()?;

    Ok(())
  }

  fn run(self) -> Result<()> {
    // References:
    // - https://github.com/python/cpython/blob/main/Lib/pty.py
    // - https://github.com/asciinema/asciinema/blob/develop/asciinema/pty_.py
    // TODO: Investigate cross-platformness
    if cfg!(target_os = "windows") {
      panic!("Not supported on windows.");
    }

    let mut command = self.command;

    // Merge environment
    for (key, val) in env::vars() {
      command.env(key, val);
    }

    // Set raw mode
    let stdin = stdin();
    let _term = RawTerm::init(stdin.as_raw_fd())?;

    // Open a pty
    let pty = openpty(None, None)?;
    let parent_stdout_dup = dup(STDOUT_FILENO)?;
    info!(parent_stdout_dup, "Duplicated parent process' stdout.");
    // dup2(pty.slave, STDOUT_FILENO)?;
    info!(
      redirected_to = pty.slave,
      "Redirected parent process' stdout to new pty."
    );

    eprintln!("Starting child process...");

    // Spawn the child
    let _child_pid = match unsafe { fork() } {
      Ok(ForkResult::Parent { child }) => {
        info!(
          child_pid = child.as_raw(),
          "Continuing execution in parent process.",
        );

        close(pty.slave)?;

        child
      }

      Ok(ForkResult::Child) => {
        setsid()?;
        close(pty.master)?;

        // Duplicate all stdio descriptors to slave
        dup2(pty.slave, STDIN_FILENO)?;
        dup2(pty.slave, STDOUT_FILENO)?;
        dup2(pty.slave, STDERR_FILENO)?;
        if pty.slave > STDERR_FILENO {
          close(pty.slave)?;
        }

        let tty_path = ttyname(STDOUT_FILENO)?;
        let tmp_fd = open(&tty_path, OFlag::O_RDWR, Mode::S_IRWXU)?;
        close(tmp_fd)?;

        exec_command(command)?;
        unreachable!()
      }

      Err(err) => bail!("Could not fork: {}", err),
    };

    // Set up fd sets and buffers for reading
    let mut read_fd_set = FdSet::new();
    let mut buf = [0; 1024];

    // Set up recording function
    let start_time = Instant::now();
    let mut last_frame_time = start_time;
    let mut record = |now: Instant, output: bool, data: &[u8]| -> Result<()> {
      let elapsed = (now - last_frame_time).as_secs_f64();
      let event_kind = (match output {
        true => EventKind::Output,
        false => EventKind::Input,
      })(data.to_vec().into());
      let event = Event(elapsed, event_kind);
      self.event_tx.send(event)?;

      last_frame_time = now;
      Ok(())
    };

    info!("Starting read loop...");
    loop {
      // Set up fdsets
      // asciinema does not capture stdin by default
      read_fd_set.clear();
      read_fd_set.insert(pty.master);
      read_fd_set.insert(STDIN_FILENO);
      // TODO: Signal file descriptor

      select(None, &mut read_fd_set, None, None, None).context("Select")?;

      // Master is ready for read, which means child process stdout has data
      // (child process stdout) -> (pty.slave) -> (pty.master)
      if read_fd_set.contains(pty.master) {
        // TODO: This line trips if the child process dies from, say, Ctrl+D.
        // Need to catch this somehow and gracefully react to it.
        // Possibly signals would be a good solution here?
        let bytes_read =
          read(pty.master, &mut buf).context("Read from master")?;
        if bytes_read == 0 {
          info!("Read 0 bytes, exiting the read loop.");
          break;
        }
        let now = Instant::now();

        let data = &buf[..bytes_read];
        debug!(bytes_read, "Read data from master.");

        // We should take this and rewrite this to the current stdout
        write(STDOUT_FILENO, data).context("Write to current STDOUT")?;

        // Also, record this
        record(now, true, data).context("Record data")?;
      }

      // Stdin is ready for read, which means input from the user
      if read_fd_set.contains(STDIN_FILENO) {
        let bytes_read =
          read(STDIN_FILENO, &mut buf).context("Read from current STDIN")?;
        let data = &buf[..bytes_read];
        debug!(bytes_read, "Read data from current STDIN.");
        let now = Instant::now();

        write(pty.master, data).context("Write to master")?;

        // Also, record this
        record(now, false, data).context("Record data")?;
      }
    }

    // Restore stdout
    fsync(STDOUT_FILENO)?;
    dup2(parent_stdout_dup, STDOUT_FILENO)?;

    Ok(())
  }
}

fn exec_command(command: Command) -> Result<()> {
  let program = CString::new(command.get_program().as_bytes())?;
  let args = command
    .get_args()
    .into_iter()
    .map(|osstr| CString::new(osstr.as_bytes()).map_err(|e| e.into()))
    .collect::<Result<Vec<_>>>()?;
  let env = command
    .get_envs()
    .into_iter()
    .filter_map(|(key, val)| {
      let val = match val {
        Some(val) => val,
        None => return None,
      };

      let mut key = key.to_os_string();
      let equals = String::from("=");
      key.extend([equals.as_ref(), val]);
      Some(CString::new(key.as_bytes()).map_err(|e| e.into()))
    })
    .collect::<Result<Vec<_>>>()?;
  execvpe(&program, &args, &env)?;
  unreachable!()
}

/*
pub fn record(args: &[&str], writer: impl Writer + Send) -> Result<()> {
  // Setup

  let forkpty_result = forkpty(None, None)?;
  let master_fd = forkpty_result.master;
  let start_time = Instant::now();

  let _set_pty_size = || -> Result<()> {
    ioctl_write_buf!(helper_write, libc::TIOCGWINSZ, 104, Winsize);
    let winsize = Winsize {
      ws_row: 24,
      ws_col: 80,
      ws_xpixel: 0,
      ws_ypixel: 0,
    };
    if isatty(STDOUT_FILENO)? {}
    unsafe { helper_write(master_fd, &[winsize]) }?;
    Ok(())
  };

  let write_stdout = |data: &[u8]| -> Result<()> {
    write(STDOUT_FILENO, &data)?;
    Ok(())
  };

  let mut master_writer = writer.clone();
  let mut handle_master_read = move |data: &[u8]| -> Result<()> {
    let elapsed = (Instant::now() - start_time).as_secs_f64();
    master_writer.write_stdout(elapsed, data)?;
    write_stdout(data)?;
    Ok(())
  };

  let write_master = |data: &[u8]| -> Result<()> {
    let mut offset = 0;
    while offset < data.len() {
      let len = write(master_fd, data)
        .map_err(|err| Error::NixWrite(err, master_fd, data.to_vec()))?;
      offset += len;
    }
    Ok(())
  };

  let mut stdin_writer = writer.clone();
  let mut handle_stdin_read = |data: &[u8]| -> Result<()> {
    write_master(data)?;
    let elapsed = (Instant::now() - start_time).as_secs_f64();
    stdin_writer.write_stdin(elapsed, data)?;
    Ok(())
  };

  // Copy fds?

  let mut copy = |signal_fd: RawFd| -> Result<()> {
    let mut fdset = FdSet::new();
    let mut buf = [0; 1024];

    loop {
      fdset.clear();
      fdset.insert(master_fd);
      fdset.insert(STDIN_FILENO);
      fdset.insert(signal_fd);

      select(None, &mut fdset, None, None, None)?;

      if fdset.contains(master_fd) {
        let len = read(master_fd, &mut buf)?;
        if len == 0 {
          fdset.remove(master_fd);
        } else {
          handle_master_read(&buf[..len])?;
        }
      } else if fdset.contains(STDIN_FILENO) {
        let len = read(STDIN_FILENO, &mut buf)?;
        if len == 0 {
          fdset.remove(STDIN_FILENO);
        } else {
          handle_stdin_read(&buf[..len])?;
        }
      } else if fdset.contains(signal_fd) {
        // TODO
      }
    }
  };

  // Fork

  let child_pid = match forkpty_result.fork_result {
    ForkResult::Parent { child } => child,
    ForkResult::Child => {
      let cstr_args: Vec<_> = args
        .into_iter()
        .map(|s| CString::new(*s).unwrap())
        .collect();
      let args: Vec<_> = cstr_args.iter().map(|s| s.as_ref()).collect();
      let cstr_env: Vec<_> = env::vars()
        .map(|(k, v)| CString::new(format!("{}={}", k, v)).unwrap())
        .collect();
      let env: Vec<_> = cstr_env.iter().map(|s| s.as_ref()).collect();
      execvpe(&args[0], &args, &env)?;
      unreachable!();
    }
  };

  let (pipe_r, pipe_w) = pipe()?;
  let mut flags = fcntl(pipe_w, FcntlArg::F_GETFL)?;
  flags |= libc::O_NONBLOCK;
  fcntl(pipe_w, FcntlArg::F_SETFL(OFlag::from_bits(flags).unwrap()))?;

  // Intercept signal handlers

  let old_handlers = set_signals(
    [
      signal_hook::SIGWINCH,
      signal_hook::SIGCHLD,
      signal_hook::SIGHUP,
      signal_hook::SIGTERM,
      signal_hook::SIGQUIT,
    ]
    .iter()
    .map(|sig| (*sig, move || eprintln!("HIT SIGNAL {:?}", *sig))),
  )?;

  // set_pty_size()?;

  {
    let _term = RawTerm::init(STDIN_FILENO)?;
    copy(pipe_r)?;
  }
  unset_signals(old_handlers)?;

  waitpid(child_pid, None)?;

  Ok(())
}

fn set_signals<I, F>(signals_list: I) -> Result<Vec<SigId>, io::Error>
where
  I: Iterator<Item = (i32, F)>,
  F: Fn() + Sync + Send + 'static,
{
  let mut old_handlers = Vec::new();
  for (sig, handler) in signals_list {
    old_handlers.push(unsafe { signal_hook::register(sig, handler) }?);
  }
  Ok(old_handlers)
}

fn unset_signals(handlers_list: Vec<SigId>) -> Result<(), io::Error> {
  for handler in handlers_list {
    signal_hook::unregister(handler);
  }
  Ok(())
}

struct RawTerm(RawFd, Termios);

impl RawTerm {
  pub fn init(fd: RawFd) -> Result<Self> {
    use nix::sys::termios::*;
    let saved_mode = tcgetattr(fd)?;
    let mut mode = saved_mode.clone();
    mode.input_flags &= !(InputFlags::BRKINT
      | InputFlags::ICRNL
      | InputFlags::INPCK
      | InputFlags::ISTRIP
      | InputFlags::IXON);
    mode.output_flags &= !OutputFlags::OPOST;
    mode.control_flags &= !(ControlFlags::CSIZE | ControlFlags::PARENB);
    mode.control_flags |= ControlFlags::CS8;
    mode.local_flags &= !(LocalFlags::ECHO
      | LocalFlags::ICANON
      | LocalFlags::IEXTEN
      | LocalFlags::ISIG);
    mode.control_chars[libc::VMIN] = 1;
    mode.control_chars[libc::VTIME] = 0;
    tcsetattr(fd, SetArg::TCSAFLUSH, &mode)?;
    Ok(RawTerm(fd, saved_mode))
  }
}

impl Drop for RawTerm {
  fn drop(&mut self) {
    tcsetattr(self.0, SetArg::TCSAFLUSH, &self.1).unwrap();
  }
}
*/
