use anyhow::Result;
use clap::{ArgAction, Parser};
use liveterm::{
  client::{
    recorder::{record_main, RecordOpts},
    stream::{stream_main, StreamOpts},
  },
  server::{server_main, ServerOpts},
};
use tracing::Level;

/// Liveterm: terminal streaming
#[derive(Parser)]
#[clap(author, version)]
struct Opt {
  #[clap(subcommand)]
  subcommand: Subcommand,

  #[clap(long, short = 'v', action = ArgAction::Count, global = true)]
  verbose: u8,
}

#[derive(Parser)]
enum Subcommand {
  /// Record terminal session.
  #[structopt(name = "record", alias = "rec")]
  Record(RecordOpts),

  /// Stream to the server
  #[structopt(name = "stream")]
  Stream(StreamOpts),

  /// Run the server.
  #[structopt(name = "server")]
  Server(ServerOpts),
}

fn main() -> Result<()> {
  let opt = Opt::parse();
  setup_logging(&opt)?;

  match opt.subcommand {
    Subcommand::Record(opts) => {
      record_main(opts)?;
    }

    Subcommand::Stream(opts) => {
      stream_main(opts)?;
    }

    Subcommand::Server(opts) => {
      server_main(opts)?;
    }
  }
  Ok(())
}

fn setup_logging(opts: &Opt) -> Result<()> {
  let file_appender = tracing_appender::rolling::hourly("logs", "liveterm.log");
  let max_level = match opts.verbose {
    0 => Level::ERROR,
    1 => Level::WARN,
    2 => Level::INFO,
    3 => Level::DEBUG,
    _ => Level::TRACE,
  };

  tracing_subscriber::fmt()
    .with_max_level(max_level)
    .with_writer(file_appender)
    .init();
  Ok(())
}
