use std::sync::atomic::{AtomicUsize, Ordering};

use anyhow::{Context, Result};
use axum::{
  extract::{
    ws::{Message as WsMessage, WebSocket},
    WebSocketUpgrade,
  },
  response::Response,
};
use dashmap::{mapref::entry::Entry, DashMap};
use futures::{future::BoxFuture, FutureExt};
use rand::{prelude::Distribution, Rng};
use serde::Deserialize;
use tokio::sync::broadcast::{self, Sender};

use crate::{
  asciicast::Header,
  message::{Message, ServerHello},
};

// TODO: Figure out a real data structure to use here
lazy_static::lazy_static! {
  pub static ref BROADCASTS: DashMap<String, Sender<Message>> = DashMap::new();
  pub static ref LEN: AtomicUsize = AtomicUsize::new(3);
}

const CAPACITY: usize = 1024;

pub async fn broadcast(ws: WebSocketUpgrade) -> Response {
  // Generate channels
  // TODO: Handle lag situation, possibly need to choose different channel
  // implementation?
  let (tx, _) = broadcast::channel(CAPACITY);

  // Allocate new room
  let mut rng = rand::thread_rng();
  let mut ct = 0;
  let room_id = loop {
    let len = LEN.load(Ordering::Relaxed);
    let room_id = (&mut rng)
      .sample_iter(RoomIdDistribution)
      .take(len)
      .collect::<Vec<_>>();
    let room_id = unsafe {
      // Distribution guarantees UTF-8-safety
      String::from_utf8_unchecked(room_id)
    };

    match BROADCASTS.entry(room_id.clone()) {
      Entry::Occupied(_) => {
        ct += 1;

        // Generous amount for now
        if ct > 10 {
          LEN.fetch_add(1, Ordering::Relaxed);
          ct = 0;
        }

        continue;
      }
      Entry::Vacant(entry) => {
        entry.insert(tx.clone());
        break room_id;
      }
    };
  };

  ws.on_upgrade(handle_websocket(room_id, tx))
}

fn deserialize_message<'d, T>(msg: &'d WsMessage) -> Result<T>
where
  T: Deserialize<'d>,
{
  Ok(match msg {
    WsMessage::Text(data) => serde_json::from_str(data)?,
    WsMessage::Binary(data) => serde_json::from_slice(data)?,
    _ => bail!("for now, protocol does not allow for other message types"),
  })
}

async fn handle_websocket_inner(
  room_id: String,
  tx: Sender<Message>,
  mut socket: WebSocket,
) -> Result<()> {
  // Step 1. Read the header
  let msg: Message = match socket.recv().await {
    Some(v) => deserialize_message(&v?)?,
    None => bail!("send a header please..."),
  };
  let _header: Header = match msg {
    Message::AsciicastHeader(header) => header,
    _ => bail!("please send header omg!!!"),
  };

  // Step 2. Respond with Hellosu
  let server_hello = Message::ServerHello(ServerHello {
    url: format!("http://192.168.0.133:8200/watch/{room_id}"),
  });
  socket
    .send(WsMessage::Text(serde_json::to_string(&server_hello)?))
    .await?;

  // Step 3. Consoooome
  while let Some(msg) = socket.recv().await {
    // TODO: Consume header

    let msg = msg.context("could not read event message")?;

    trace!(?msg, "parsed event message");

    let msg: Message = match &msg {
      WsMessage::Text(data) => serde_json::from_str(data),
      WsMessage::Binary(data) => serde_json::from_slice(data),
      _ => continue,
    }
    .context("could not parse event message from json")?;

    // Ignoring this for now cus it'll error if there's no receivers
    tx.send(msg);
  }

  Ok(())
}

fn handle_websocket<'a>(
  room_id: String,
  tx: Sender<Message>,
) -> impl FnOnce(WebSocket) -> BoxFuture<'a, ()> {
  move |socket: WebSocket| {
    async {
      match handle_websocket_inner(room_id, tx, socket).await {
        Ok(_) => {}
        Err(err) => {
          eprintln!("Fuck! {err:?}");
          eprintln!("{err}");
        }
      }
    }
    .boxed()
  }
}

// Cribbed from https://rust-random.github.io/rand/rand/distributions/struct.Alphanumeric.html
struct RoomIdDistribution;

impl Distribution<u8> for RoomIdDistribution {
  fn sample<R: Rng + ?Sized>(&self, rng: &mut R) -> u8 {
    const RANGE: u32 = 26 + 26 + 10;
    const ALLOWED_CHARS: &[u8] = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ\
                abcdefghijklmnopqrstuvwxyz\
                0123456789\
                -";
    loop {
      let var = rng.next_u32() >> (32 - 6);
      if var < RANGE {
        return ALLOWED_CHARS[var as usize];
      }
    }
  }
}
